@@ -233,6 +233,9 @@ static struct alisp_object * incref_tree(struct alisp_instance *instance, struct
 	return incref_object(instance, p);
 }
 
+/* Function not used yet. Leave it commented out until we actually use it to
+ * avoid compiler complaints */
+#if 0
 static struct alisp_object * incref_tree_explicit(struct alisp_instance *instance, struct alisp_object * p, struct alisp_object * e)
 {
 	if (p == NULL)
@@ -250,6 +253,7 @@ static struct alisp_object * incref_tree_explicit(struct alisp_instance *instanc
 		return incref_object(instance, p);
 	return p;
 }
+#endif
 
 static void free_objects(struct alisp_instance *instance)
 {
@@ -1708,7 +1712,7 @@ static struct alisp_object * F_princ(struct alisp_instance *instance, struct ali
 			delete_tree(instance, p1);
 		p1 = eval(instance, car(p));
 		if (alisp_compare_type(p1, ALISP_OBJ_STRING))
-			snd_output_printf(instance->out, p1->value.s);
+			snd_output_printf(instance->out, "%s", p1->value.s);
 		else
 			princ_object(instance->out, p1);
 		n = cdr(p);
@@ -2965,7 +2969,7 @@ struct intrinsic {
 	struct alisp_object * (*func)(struct alisp_instance *instance, struct alisp_object * args);
 };
 
-static struct intrinsic intrinsics[] = {
+static const struct intrinsic intrinsics[] = {
 	{ "!=", F_numneq },
 	{ "%", F_mod },
 	{ "&check-memory", F_check_memory },
@@ -3252,7 +3256,7 @@ int alsa_lisp(struct alisp_cfg *cfg, struct alisp_instance **_instance)
 	else
 		alsa_lisp_free(instance); 
 	
-	return 0;
+	return retval;
 }
 
 void alsa_lisp_free(struct alisp_instance *instance)
